/**
 * the network chunk-API
 * 
 * 
 */

#include <sys/types.h>
#include <sys/mman.h>
#include <sys/stat.h>

#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>

#include <stdio.h>
#include <errno.h>
#include <string.h>

#include "chunk.h"

chunkset *chunk_init() {
	chunkset *c;
	
	c = malloc(sizeof(chunkset));
	
	c->chunk = NULL;
	c->size = 0;
	c->used = 0;
	
	return c;
}

int chunk_prepare_append(chunkset *c) {
	size_t i;
	if (0 == c->size) {
		c->size = 4;
		c->chunk = malloc(c->size * sizeof(chunk *));
		
		for (i = 0; i < c->size; i++) {
			c->chunk[i] = malloc(sizeof(chunk));
			c->chunk[i]->type = UNUSED_CHUNK;
			c->chunk[i]->offset = 0;
			c->chunk[i]->length = 0;
			c->chunk[i]->written = 0;
			c->chunk[i]->b = buffer_init();
		}
	} else if (c->used == c->size) {
		/* resize */
		
		c->size += 4;
		c->chunk = realloc(c->chunk, c->size * sizeof(chunk *));
		
		for (i = c->used; i < c->size; i++) {
			c->chunk[i] = malloc(sizeof(chunk));
			c->chunk[i]->type = UNUSED_CHUNK;
			c->chunk[i]->offset = 0;
			c->chunk[i]->length = 0;
			c->chunk[i]->written = 0;
			c->chunk[i]->b = buffer_init();
		}
	}
	return 0;
}

int chunk_append_file(chunkset *c, buffer *fn, off_t offset, off_t len) {
	if (len == 0) return 0;
	
	chunk_prepare_append(c);
	
	c->chunk[c->used]->type = FILE_CHUNK;
	
	buffer_copy_string_len(c->chunk[c->used]->b, fn->ptr, fn->used - 1);
	c->chunk[c->used]->offset = offset;
	c->chunk[c->used]->length = len;
	c->chunk[c->used]->written = 0;
	
	c->used++;
	
	return 0;
}

int chunk_insert_buffer(chunkset *c, int pos, buffer *mem) {
	if (mem->used == 0) return 0;
	
	c->chunk[pos]->type = MEM_CHUNK;
	
	buffer_copy_string_len(c->chunk[pos]->b, mem->ptr, mem->used - 1);
	
	c->chunk[pos]->written = 0;
	
	return 0;
}

int chunk_append_buffer(chunkset *c, buffer *mem) {
	if (mem->used == 0) return 0;
	
	chunk_prepare_append(c);
	
	chunk_insert_buffer(c, c->used, mem);
	
	c->used++;
	
	return 0;
}

int chunk_insert_mem(chunkset *c, int pos, const char *mem, size_t len) {
	if (len == 0) return 0;
	
	c->chunk[pos]->type = MEM_CHUNK;
	
	buffer_copy_string_len(c->chunk[pos]->b, mem, len - 1);
	
	c->chunk[pos]->written = 0;
	
	return 0;
}

int chunk_append_mem(chunkset *c, const char * mem, size_t len) {
	if (len == 0) return 0;
	
	chunk_prepare_append(c);
	
	chunk_insert_mem(c, c->used, mem, len);
	
	c->used++;
	
	return 0;
}

int chunk_prepend_buffer(chunkset *c, buffer *mem) {
	if (mem->used == 0) return 0;
	
	chunk_prepare_append(c);
	
	/* prepare the first chunk */
	if (c->used) {
		int i;
		chunk *d;
		
		d = c->chunk[c->used];
		for (i = c->used; i > 0; i--) {
			c->chunk[i] = c->chunk[i - 1];
		}
		c->chunk[0] = d;
	}
	
	chunk_insert_buffer(c, 0, mem);
	
	c->used++;
	
	return 0;
}

buffer * chunk_get_prepend_buffer(chunkset *c) {
	chunk_prepare_append(c);
	
	/* prepare the first chunk */
	if (c->used) {
		int i;
		chunk *d;
		
		d = c->chunk[c->used];
		for (i = c->used; i > 0; i--) {
			c->chunk[i] = c->chunk[i - 1];
		}
		c->chunk[0] = d;
	}
	
	c->chunk[0]->type = MEM_CHUNK;
	c->chunk[0]->written = 0;
	
	c->used++;
	
	return c->chunk[0]->b;
}

buffer *chunk_get_append_buffer(chunkset *c) {
	chunk_prepare_append(c);
	
	c->chunk[c->used]->type = MEM_CHUNK;
	c->chunk[c->used]->written = 0;
	
	return c->chunk[c->used++]->b;
}


void chunk_free(chunkset *c) {
	size_t i;
	
	for (i = 0; i < c->size; i++) {
		if (c->chunk[i]->b) buffer_free(c->chunk[i]->b);
		
		free(c->chunk[i]);
	}
	
	if (c->chunk) free(c->chunk);
	free(c);
}


size_t chunkset_length(chunkset *c) {
	size_t i, len = 0;
	
	for (i = 0; i < c->used; i++) {
		switch (c->chunk[i]->type) {
		case MEM_CHUNK:
			len += c->chunk[i]->b->used ? c->chunk[i]->b->used - 1 : 0;
			break;
		case FILE_CHUNK:
			len += c->chunk[i]->length;
			break;
		default:
			break;
		}
	}
	
	return len;
}

size_t chunkset_written(chunkset *c) {
	size_t i, len = 0;
	
	for (i = 0; i < c->used; i++) {
		switch (c->chunk[i]->type) {
		case MEM_CHUNK:
		case FILE_CHUNK:
			len += c->chunk[i]->written;
			break;
		default:
			break;
		}
	}
	
	return len;
}

#ifdef CHUNK_API_TEST
int main() {
	chunkset *c;
	buffer *b;
	
	c = chunk_init();
	
	chunk_append_file(c, "server.c", 0, 10);
	chunk_append_file(c, "server.c", 10, 10);
	chunk_append_file(c, "server.c", 20, 10);
	
	network_write_chunkset(2, c);
	c->used = 0;
	
	b = buffer_init();
	buffer_copy_string(b, "\ntest string mit vielen Zeichen\n");
	chunk_append_buffer(c, b);
	
	network_write_chunkset(2, c);
	c->used = 0;
	
	chunk_append_file(c, "server.c", 0, 10);
	buffer_copy_string(b, "\ntest string mit vielen Zeichen\n");
	chunk_append_buffer(c, b);
	chunk_append_file(c, "server.c", 10, 10);
	chunk_append_file(c, "server.c", 20, 10);
	chunk_append_file(c, "server.c", 50, 40);
	
	network_write_chunkset(2, c);
	c->used = 0;
	
	chunk_free(c);
	
	return 0;
}
#endif
