#define _GNU_SOURCE

#include <sys/types.h>
#include <sys/socket.h> /* inet_ntop */

#include <netinet/in.h>

#include <arpa/inet.h>

#include <errno.h>
#include <fcntl.h>
#include <time.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>

#include <stdarg.h>
#include <stdio.h>

#include "log.h"
#include "array.h"

#ifndef O_LARGEFILE
# define O_LARGEFILE 0
#endif

int log_error_open(server *srv, const char *logfile) {
	if (logfile) {
		if (-1 == (srv->log_error_fd = open(logfile, O_APPEND | O_WRONLY | O_CREAT | O_LARGEFILE, 0644))) {
			char *msg = "opening error-log failed: ";
			write(2, msg, strlen(msg));
			msg = strerror(errno);
			write(2, msg, strlen(msg));
			write(2, "\n", 1);
			return -1;
		}
		
		close(2);
	} else {
		srv->log_error_fd = 2;
	}
	
	log_error_write(srv, __FILE__, __LINE__, "s", "server started");
	
	return 0;
}

int log_error_close(server *srv) {
	log_error_write(srv, __FILE__, __LINE__, "s", "server stopped");
	
	if (srv->log_error_fd != 2) close(srv->log_error_fd);
	
	return 0;
}

int log_error_write(server *srv, const char *filename, unsigned int line, const char *fmt, ...) {
	va_list ap;
	
	/* cache the generated timestamp */
	if (srv->cur_ts != srv->last_generated_debug_ts) {
		buffer_prepare_copy(srv->ts_debug_str, 255);
		strftime(srv->ts_debug_str->ptr, srv->ts_debug_str->size - 1, "%Y-%m-%d %H:%M:%S", localtime(&(srv->cur_ts)));
		srv->ts_debug_str->used = strlen(srv->ts_debug_str->ptr) + 1;
		
		srv->last_generated_debug_ts = srv->cur_ts;
	}

	buffer_copy_string_buffer(srv->error_log, srv->ts_debug_str);
	buffer_append_string_len(srv->error_log, ": (", 3);
	buffer_append_string(srv->error_log, filename);
	buffer_append_string_len(srv->error_log, ".", 1);
	buffer_append_long(srv->error_log, line);
	buffer_append_string_len(srv->error_log, ") ", 2);
	
	
	for(va_start(ap, fmt); *fmt; fmt++) {
		int d;
		char *s;
		buffer *b;
		
		switch(*fmt) {
		case 's':           /* string */
			s = va_arg(ap, char *);
			buffer_append_string(srv->error_log, s);
			buffer_append_string_len(srv->error_log, " ", 1);
			break;
		case 'b':           /* buffer */
			b = va_arg(ap, buffer *);
			buffer_append_string_buffer(srv->error_log, b);
			buffer_append_string_len(srv->error_log, " ", 1);
			break;
		case 'd':           /* int */
			d = va_arg(ap, int);
			buffer_append_long(srv->error_log, d);
			buffer_append_string_len(srv->error_log, " ", 1);
			break;
		case 'S':           /* string */
			s = va_arg(ap, char *);
			buffer_append_string(srv->error_log, s);
			break;
		case 'B':           /* buffer */
			b = va_arg(ap, buffer *);
			buffer_append_string_buffer(srv->error_log, b);
			break;
		case 'D':           /* int */
			d = va_arg(ap, int);
			buffer_append_long(srv->error_log, d);
			break;
		case '(':
		case ')':
		case '<':	
		case '>':
		case ',':
		case ' ':
			buffer_append_string_len(srv->error_log, fmt, 1);
			break;
		}
	}
	va_end(ap);
	
	buffer_append_string_len(srv->error_log, "\n", 1);
	
	write(srv->log_error_fd ? srv->log_error_fd : 2, srv->error_log->ptr, srv->error_log->used - 1);
	
	return 0;
}

